function [g_low,g_high]=periodicL2gain(persys,verbose)
%Computes the L2 gain of periodic linear system by using the iterative
%algorithm published in
% M. Cantoni, H. Sandberg: Computing the L2 gain for linear
%  periodic continuous-time systems. Automatica 45, pp. 783-789, 2009.
%  (Version 4)





if nargin==0 %demo
    clear all; close all;
    persys=@demosys;
    verbose=1;
end;
if verbose, tic; end;
h=persys([]);
[A,B,C,D]=persys(0);
nx=size(A,1); nw=size(B,2); nz=size(C,1);
options = odeset('RelTol',1e-6,'AbsTol',1e-9);
epps=0.0001;

%switching off annoying warnings of ode45
warning('off','MATLAB:ode45:IntegrationTolNotMet');


%--Step I. ----------------------------------------------------------------
[~,gl]=fminbnd(@(t)-Dnrm(persys,t),0,h);
gl=max(-gl,1e-4);

if verbose, fprintf('\nComputing bounds [g_low,g_high] on ||P^||_inf by using Algorithm 2.\n'); end;

g=gl; g_high=[]; g_low=gl; gap=1;
while gap>epps
    
    if verbose, fprintf('--------[%8.5f, %8.5f]\n',g_low,g_high); end;
    
    
    [T,Zvec]=computeZ(g,persys,nx,nz,nw,h,options);
    if abs(T(end))==0 && norm(Zvec(end,:))<Inf,
        % Bounded Solution => ||D^||<g
        Z0=reshape(Zvec(end,:),nx,nx);
        Z0=(Z0+Z0')/2;
        %Cg=chol(Z0);
        Cg=sqrtm(Z0);
        
        [~,Xvec]=computeX(g,T,Zvec,persys,nx,nz,nw,h,options);
        Xh=reshape(Xvec(end,:),nx,nx);
        Ag=Xh;
        
        
        [~,Yvec]=computeY(g,persys,nx,nz,nw,h,options);
        Yh=reshape(Yvec(end,:),nx,nx);
        Yh=(Yh+Yh')/2;
        %Bg=chol(Yh)';
        Bg=sqrtm(Yh)';
        
        DtSys=ss(Ag,Bg,Cg,0,-1);
        nrm=norm(DtSys,inf,1e-9);
        
       
        
        if nrm<1, %||D^||<||P^||_inf<g
            g_high=g; g=(g_high+g_low)/2; gap=g_high-g_low;
        else %||D^||< g <||P^||_inf
            g_low=g;
            if isempty(g_high),
                g=2*g;
            else
                g=(g_high+g_low)/2; gap=g_high-g_low;
            end;
        end;
    else
        % Unbounded Solution: g < ||D^||
        if isempty(g_high)
            g=2*g;
        else
            g_low = g;g=(g_high+g_low)/2; gap=g_high-g_low;
        end
    end
    
end;
if verbose, fprintf('finally:[%8.5f, %8.5f]\n',g_low,g_high); end;

%Restoring warnings
warning('on','MATLAB:ode45:IntegrationTolNotMet');
if verbose, toc; end;




function [T,Zvec]=computeZ(g,persys,nx,nz,nw,h,options)
%Computing Z(t) in the interval [h,0] by using Eq. (17)
[T,Zvec]=ode45(@(t,Zvec) reshape(Zdotfun(t,reshape(Zvec,nx,nx),persys,g,nw,nz),nx^2,1),[h,0],zeros(nx^2,1),options);

function Zdot=Zdotfun(t,Z,persys,g,nw,nz)
[A,B,C,D]=persys(t);
R=g^2*eye(nw)-D'*D;
S=g^2*eye(nz)-D*D';
H22=A+B*(R\D')*C;
Zdot=-H22'*Z-Z*H22-g*Z*B*(R\B')*Z-g*C'*(S\C);





function [T,Xvec]=computeX(g,Tz,Zvec,persys,nx,nz,nw,h,options)
%Computing X(t) in the interval [0,h] by using Eq. (18)
[T,Xvec]=ode45(@(t,Xvec) reshape(Xdotfun(t,reshape(Xvec,nx,nx),persys,g,nw,nz,Tz,Zvec),nx^2,1), [0 h],reshape(eye(nx),nx^2,1),options);

function Xdot=Xdotfun(t,X,persys,g,nw,nz,Tz,Zvec)
[A,B,C,D]=persys(t);
R=g^2*eye(nw)-D'*D;
S=g^2*eye(nz)-D*D';
H22=A+B*(R\D')*C;
Xdot=(H22+g*B*(R\B')*Zfun(Tz,Zvec,t))*X;





function [T,Yvec]=computeY(g,persys,nx,nz,nw,h,options)
%Computing Y(t) in the interval [0,h] by using Eq. (19)
[T,Yvec]=ode45(@(t,Yvec) reshape(Ydotfun(t,reshape(Yvec,nx,nx),persys,g,nw,nz),nx^2,1), [0 h],zeros(nx^2,1),options);

function Ydot=Ydotfun(t,Y,persys,g,nw,nz)
[A,B,C,D]=persys(t);
R=g^2*eye(nw)-D'*D;
S=g^2*eye(nz)-D*D';
H22=A+B*(R\D')*C;
Ydot=(H22*Y+Y*H22'+g*Y*C'*(S\C)*Y+g*B*(R\B'));




function Z=Zfun(T,Zvec,t)
%To determine X(t) by forward integration we need Z(t), which is computed
%by backward integration (Eqs. 18 and 17). Having integrated Eq. 17 we get
%Z(t) at discrete points (output of ode45)). The 'missing values'  are
%computed by interpolation.

% XXX PJS: Call to interp1 is slow since it does extra error checking, etc.
% n=size(Zvec,2);
% Z=zeros(n,1);
% for i=1:n
%     Z(i)=interp1(T,Zvec(:,i),t,'linear');
% end;
% Z=reshape(Z,n/2,n/2);

% Direct implementation of linear interpolation is faster
Nt = length(T);
i = min(find(t>=T));
if i>1
    alpha = (t-T(i)) / (T(i-1)-T(i));
    Z = (1-alpha)*Zvec(i,:)+alpha*Zvec(i-1,:);
else
    Z = Zvec(1,:);
end
n=size(Zvec,2); sqrtn=sqrt(n);
Z=reshape(Z,sqrtn,sqrtn);


function res=Dnrm(persys,t)
%Norm function, which works with vector argument. (Required by fminbnd)
res=zeros(1,length(t));
for i=1:length(t)
    [~,~,~,D]=persys(t(i));
    res(i)=norm(D);
end;



function [A,B,C,D]=demosys(t)
if isempty(t), A=pi*100; return; end;
t=t/100;
A=[-1-sin(2*t)^2 2-0.5*sin(4*t);-2-0.5*sin(4*t) -1-cos(2*t)^2];
B=[0;1];
C=[1 1];
D=1;



